#include "MyQOpenGLWindow.h"

#include "Renderer.h"
#include "Animator.h"
#include <QMouseEvent>
#include <QScreen>
#include <QApplication>
#include <QElapsedTimer>
#include <QTimer>
#include <cassert>
#include <iostream>
#include <algorithm> // std::copy()

MyQOpenGLWindow::MyQOpenGLWindow() : QOpenGLWindow()
{
    connect(this, &QWindow::screenChanged, this,
            &MyQOpenGLWindow::updateDevicePixelRatio);

    m_fpsTimer = new QTimer;
    connect(m_fpsTimer, &QTimer::timeout, this, &MyQOpenGLWindow::checkFPS);
    m_fpsTimer->start(m_fpsTimeInMs);

    m_elapsedTimer = new QElapsedTimer;
    m_tickTimer = new QTimer;
    connect(m_tickTimer, &QTimer::timeout, this, &MyQOpenGLWindow::animate);
    m_tickTimer->start(1000.0f / 60.f); // 60Hz
}

MyQOpenGLWindow::~MyQOpenGLWindow()
{
    delete m_fpsTimer;
    delete m_tickTimer;
    delete m_elapsedTimer;
}

void MyQOpenGLWindow::updateDevicePixelRatio(QScreen* s)
{
    Q_UNUSED(s);
    // We cannot just call resize(size()) to trigger resizeGL() because Qt
    // will detect no change in size and will not propagate the change so we
    // call resizeGL() directly.
    QSize currentSize = size();
    resizeGL(currentSize.width(), currentSize.height());
}

Renderer* MyQOpenGLWindow::renderer() const
{
    return m_renderer;
}

void MyQOpenGLWindow::setRenderer(Renderer* r)
{
    m_renderer = r;
    assert("post: is_set" && renderer() == r);
}

Animator* MyQOpenGLWindow::animator() const
{
    return m_animator;
}

void MyQOpenGLWindow::setAnimator(Animator* a)
{
    m_animator = a;
    if (m_animator != nullptr) {
        m_elapsedTimer->start();
        m_animator->setTime(milliSecondsToSeconds(m_elapsedTimer->elapsed()));
    }
    assert("post: is_set" && animator() == a);
}

void MyQOpenGLWindow::initializeGL()
{
    QSurfaceFormat f = format();
    std::cout << "format: stereo=" << f.stereo()
              << " alphaBufferSize=" << f.alphaBufferSize()
              << " samples=" << f.samples()
              << " depthBufferSize=" << f.depthBufferSize()
              << " stencilBufferSize=" << f.stencilBufferSize() << std::endl;

    if (m_renderer != nullptr) {
        m_renderer->initializeGL();
    }
}

void MyQOpenGLWindow::paintGL()
{
    if (m_renderer != nullptr) {
        m_renderer->paintGL();
    } else {
        std::cerr << "MyQOpenGLWindow::paintGL(): No renderer" << std::endl;
    }
}

void MyQOpenGLWindow::resizeGL(int w_points, int h_points)
{
    // Contrarily to Q(Open)GLWidget::resizeGL(), there is no guarantee that
    // makeCurrent was called before entering this method.
    makeCurrent();
    double ratio = devicePixelRatio();
    int w_pixels = static_cast<int>(w_points * ratio);
    int h_pixels = static_cast<int>(h_points * ratio);

    if (m_renderer != nullptr) {
        m_renderer->resizeGL(w_pixels, h_pixels);
    }
    doneCurrent();
}

void MyQOpenGLWindow::checkFPS()
{
    if (m_renderer != nullptr) {
        int counter = m_renderer->paintGLCounter();
        m_renderer->resetPaintGLCounter();
        std::cout << "FPS=" << (counter / (m_fpsTimeInMs / 1000.0f))
                  << " frametime=" << (m_fpsTimeInMs / counter) << " ms"
                  << std::endl;
    }
}

void MyQOpenGLWindow::animate()
{
    //  std::cout << "void MyQOpenGLWindow::animate()" << std::endl;
    if (m_animator != nullptr) {
        m_animator->animate(milliSecondsToSeconds(m_elapsedTimer->elapsed()));
        update(); // from the painter superclass
        // requestUpdate(); // or from the QWindow superclass
    }
}

float MyQOpenGLWindow::milliSecondsToSeconds(float milli) const
{
    return milli / 1000.0f;
}

